/* **************************************************** */
/* file bp.cpp:  contains the main program and network  */
/*               creation routines.                     */
/*                                                      */
/* Copyright (c) 1990-98 by Donald R. Tveter            */
/*                                                      */
/* **************************************************** */

#ifdef INTEGER
#include "ibp.h"
#else
#include "bp.h"
#endif

/* built-in C functions */

extern int rand();
extern void srand(unsigned);

#ifdef INTEGER
extern short scale(REAL);
extern REAL unscale(short);
#endif

extern void maze(int,int,int);
extern void evalr(int,int,int);
extern void initialize_etas(WTTYPE);
extern void nullpatterns(int);
extern void findendofpats(LAYER *);
extern void nextpat(int);
extern WTTYPE rdr(int,REAL,char);
extern int pg(FILE *, char *);
extern printoutunits(int, LAYER*,int,REAL,char,char);
extern int readch();
extern eval(int,int);
extern int readpats(int,int);
extern int evalone(int,int,int,int);
extern int printstats(FILE *,int,int,DATA [2][2]);
extern void popfile();
extern int pushfile(char *, int);
extern REAL readreal(int,REAL,char);
extern int readint(int,int,char);
extern char *readstr();
extern int loadpat();
extern int run(int,int);
extern void texterror();
extern void clear();
extern void kick(WTTYPE,WTTYPE);
extern void restoreweights();
extern void saveweights();
extern void parameters(FILE *);
extern void printweights(UNIT *, int);
extern void resetpats(int);

char sysstr[129];     /* string passed to OS with ! command */
char path[129];       /* path for the new directory */

int bufferend;        /* index of last character in input line */
int bufferptr;        /* position of next character in buffer */
char buffer[BUFFSIZE];/* holds contents of one input line */
char trfiles[RXSTRSIZE];/* data file name(s) */
int ch;               /* general purpose character variable */
FILE *copy;           /* file pointer to copy file */
char copyflag;        /* + for copying, - for no copy */
int copyfilecount;    /* to number the copy files */
char *copyfilename;
char cmdfilename[FILENAMESIZE];
jmp_buf endofccmd;    /* to save state in case of a SIGINT */
FILE *dataf;          /* file for original data */
char *datafile;       /* copy of the data file name saved here */

WTTYPE dbdeta;        /* the initial eta value for the DBD method */
WTTYPE kappa;         /* the DBD learning parameter */
WTTYPE decay;         /* the decay parameter for the DBD method */
WTTYPE etamax;        /* the maximum eta for the DBD method */
WTTYPE theta1;        /* the DBD parameter */
WTTYPE theta2;        /* 1 - theta1 */
WTTYPE noise;         /* noise parameter for dbd */

char deriv;           /* flags type of derivative to use */
char echo;            /* controls echoing of characters during input */
char emptystring;     /* for unused string values */

WTTYPE eta;           /* basic learning rate */
WTTYPE eta2;          /* learning rate for lower layers */
WTTYPE alpha;         /* momentum term */

char recurinit;       /* intialize input and output at start of training */
int extraconnect;     /* flags the use of connections between */
                      /* non-adjacent layers */
char *inputfile;      /* name of file to take extra commands from */
FILE *filestack[MAXFILES];  /* allows for nested reads from files */
int filetimes[MAXFILES];    /* number of times to read file */
int filestackptr;     /* has the index of the current file */
int format[MAXFORMAT];/* each value in format indicates where to put */
                      /* a blank for compressed output mode or a */
                      /* carriage return for real output */
int timestoread;      /* number of times to read an input file */
char incrementseed;   /* + advances seed, - resets to first value */
char informat;        /* controls format to read numbers */
WTTYPE initrange;     /* initial range of random weights */
WTTYPE kicksize;      /* the minimum size weights that are affected */
WTTYPE kickrange;     /* the range weights are initialized to */
char wtinitroutine;   /* flags weight initialization algorithm */

long iotime;          /* time used up printing */
short nlayers;        /* number of layers in network */
int ioconnects;       /* flags existence of input/output connections */
int originallypcc;    /* saves kind of net to start with */

LAYER *last;          /* has address of the output layer */
LAYER *start;         /* has address of the input layer */
int stmsize;          /* number of h layer units in a recurrent net */

int lastadd;          /* iteration where last node was added */
REAL lastadderr;      /* error at the last hidden unit addition */
int lastprint;        /* last iteration pattern responses printed */
int lastsave;         /* last time weights were saved */
int lineno;           /* counts lines for paging */
int maxhidden;        /* maximum number of hidden units allowed */
int maxiter;          /* maximum iterations when not benchmarking */
char netbuild;        /* the type of net to build */
char offby1;          /* + means out of date stats for training set */
char outformat;       /* controls format to print output */
char outstr[OUTSTRSIZE]; /* the output string */
int pagesize;         /* size of page for pg */
char runningflag;     /* +/- to print the "running . . ." message */
REAL pasterror;       /* error from window iterations ago */
int pccnet;           /* flags pseudo-cascade-correlation net */
char probtype;        /* flags general or classification pattern format */
char debugoutput;     /* turns on any debugging code I need */
int prevnpats;        /* previous number of patterns, initially 0 */
int printrate;        /* printrate when not benchmarking */
WTTYPE unknown;       /* value for x in compressed input */

REAL qpdecayh;        /* the quickprop weight decay hidden layer */
REAL qpdecayo;        /* the quickprop weight decay output layer */
WTTYPE qpeta;         /* the quickprop eta */
WTTYPE qpnoise;       /* quickprop noise (integer version only) */
WTTYPE mu;            /* the quickprop acceleration factor */
char qpslope;         /* flags using slope in quickprop all the time */

int readerror;        /* flags an error in reading a value */
int readingpattern;   /* flags reading pattern state */
WTTYPE classoff;      /* target value for class off value */
WTTYPE classon;       /* target value for class on value */
char ringbell;        /* flag to ring bell when finished */
int right;            /* number of training patterns learned */
DATA s[2][2];         /* contains stats on pattern learning */
DATA rs[2][2];        /* stats on recurrent pattern learning */
int saverate;         /* rate at which to save weights */
char saveonminimum;   /* flags saving weights on minimum of test set */
REAL minimumsofar;    /* lowest test set error so far */
unsigned seed;        /* seed for generating random weights */
SEEDNODE *seedstart;  /* the list of user defined seeds */
#ifdef LOOKUP
WTTYPE sigmoid[7808]; /* table to look up the sigmoid's value */
#endif

WTTYPE stdthresh;     /* the bias WEIGHT value */
WTTYPE hbiasact;      /* hidden layer bias activation VALUE */
WTTYPE obiasact;      /* output layer bias activation VALUE */
char biasset;         /* flags the setting of the bias WEIGHT */
char summary;         /* flags summary output mode */
int testpat;          /* pattern to skip when benchmarking; else 0 */
char *testfile;       /* file to take test patterns from */
WTTYPE toler;         /* tolerance based on targets */
REAL toloverall;      /* tolerance based on average error */
WTTYPE toosmall;      /* weights smaller than toosmall were removed */
#ifdef INTEGER
INT32 totaldiff;      /* totals errors to find average error per unit */
#else
REAL totaldiff;
#endif
int totaliter;        /* counts total iterations for the program */
char *trainfile;      /* file to take training patterns from */
REAL trigger;         /* trigger slope for dynamic node creation */
char update;          /* flags type of update rule to use */
char up_to_date_stats;/* + does an extra forward pass after update */
int window;           /* window over which trigger slope is computer */
int wrong;            /* number of training patterns unlearned */
char wtfilename[FILENAMESIZE]="weights"; /* whole name for weight file */
char *wtfile;         /* base file name for weights file */
int wtfilecount;      /* counts weight files as they are written */
char wtformat;        /* controls format to save and restore weights */
#ifdef INTEGER
char wtlimithit;      /* flags whether the limit has been hit */
#endif
int wttotal;          /* number of weights in the network */
int wtsinuse;         /* number of weights in use */
char zeroderiv;       /* flags zero deriviative when using A ap */

char ah;              /* hidden layer activation function */
char ao;              /* output layer activation function */
WTTYPE Dh;            /* hidden layer D */
WTTYPE Do;            /* output layer D */

REAL poft;            /* temporal difference previous value */

int ninput;          // number of input units
int n1hidden;        // number of first hidden layer units
int n2hidden;        // number of second hidden layer units
int noutput;         // number of output units

/* Tcl/Tk GUI Stuff */

Tcl_Interp *interp;
int windwidth;       // width of window
int windheight;      // height of window
int upperleftx;      // upper left corner x
int upperlefty;      // upper left corner y
int backgroundcolor; // background color
int textcolor;       // text color
int currentfont;     // current font number
int maxlines;        // maximum lines in canvas
char errormessages;  // flags type of error messages displayed
char tkbuf[OUTSTRSIZE]; // collects characters for the tk canvas
int tkbufptr;
int sigint;             // flags a sigint
ClientData clientdata;  // need for Tcl_AsyncCreate
int breakloop;


void error(char *s)
{
 char str[100];
 if (errormessages == 'b' || errormessages == 't') pg(NULL,s);
 if (errormessages == 'b' || errormessages == 'w')
  {
   sprintf(str,"error \"%s\"",s);
   Tcl_Eval(interp,str);
  };
}

/* given a layer no. and unit no. locateunit returns the address */
UNIT *locateunit(int layerno,int unitno)
{int i;
 UNIT *u;
 LAYER *layer;
 
if (layerno < 0 || layerno > nlayers)
 {
  error("bad layer number\n");
  return(NULL);
 };
layer = start;
for(i=1;i<=(layerno-1);i++) layer = layer->next;
u = (UNIT *) layer->units;
while (u != NULL && u->unitnumber != unitno) u = u->next;
if (u == NULL)
 {
  sprintf(outstr,"there is no unit %3d in layer %3d\n",unitno,layerno);
  error(outstr);
 };
return(u);     
}

#ifdef SYMMETRIC

INT32 wtaddress(i,j,biasunit,type,size) /* Returns the address of a */
int i,j;                                /* weight (1), olddw (2),   */
int biasunit;                           /* eta (3), total (4),      */
int type;                               /* or slope (5)             */
int size;                               /* One is created if it     */
                                        /* doesn't already exist.   */
{ int k;
  INT32 addr;
  UNIT *u;
  WTNODE *w;

if (biasunit) addr = (INT32) malloc(size);
else if (j >= i) addr = (INT32) malloc(size);
else /* the item already exists, so find its address */
 {
  u = locateunit(2,j);
  w = (WTNODE *) u->wtlist;
  k = 1;
  while (k < i)
   {
    w = w->next;
    k = k + 1;
   };
  if (type == 1) addr = (INT32) w->weight;
  else if (type == 2) addr = (INT32) w->olddw;
  else if (type == 3) addr = (INT32) w->eta;
  else if (type == 4) addr = (INT32) w->total;
  else if (type == 5) addr = (INT32) w->slope;
  else error("bad type in wtaddress\n");
 };
return(addr);
}

void setweight(WTNODE *w,short i,short j,int biasunit)
/* set initial values in w */
{WTTYPE *s;

s = (WTTYPE *) wtaddress(i,j,biasunit,1,WTSIZE);
*s = 0;
w->weight = s;
s = (WTTYPE *) wtaddress(i,j,biasunit,2,WTSIZE);
*s = 0;
w->olddw = s;
s = (WTTYPE *) wtaddress(i,j,biasunit,3,WTSIZE);
*s = eta;
w->eta = s;
s = (WTTYPE *) wtaddress(i,j,biasunit,5,WTSIZE);
*s = 0;
w->slope = s;
w->inuse = 1;
#ifdef INTEGER
w->total = (INT32 *) wtaddress(i,j,biasunit,4,sizeof(INT32));
#else
w->total = (REAL *) wtaddress(i,j,biasunit,4,sizeof(REAL));
#endif
}

#else

void setweight(WTNODE *w,short i,short j,int biasunit) /* set initial values in w */
{
w->weight = 0;
w->olddw = 0;
w->slope = 0;
w->eta = dbdeta;
if (biasunit) w->inuse = 2; else w->inuse = 1;
}

#endif

LAYER *mklayer(LAYER *prevlayer,int n)
// creates a layer of n units, pointers and weights back to the units in
// the previous layer and links this new layer into the list of layers 
{UNIT *front, *p, *q, *bias, *prev, *ptr;
 WTNODE *wfront, *wprev, *w;
 LAYER *lptr;
 int i, j, count;

/* make a list of nodes in this layer */

count = 1;
front = (UNIT *) malloc(sizeof(UNIT));
front->unitnumber = count;
front->layernumber = nlayers;
front->oj = 0;
front->tj = 0;
front->translate = 0;
front->userscale = scale(1.0);
front->wtlist = NULL;
prev = front;
for(i=1;i<n;i++)
 {
  count = count + 1;
  ptr = (UNIT *) malloc(sizeof(UNIT));
  prev->next = ptr;
  ptr->unitnumber = count;
  ptr->layernumber = nlayers;
  ptr->wtlist = NULL;
  ptr->translate = 0;
  ptr->userscale = scale(1.0);
  front->oj = 0;
  front->tj = 0;
  prev = ptr;
 };
prev->next = NULL;

/* make a LAYER node to point to this list of units */

lptr = (LAYER *) malloc(sizeof(LAYER));
lptr->unitcount = n;
lptr->initialcount = n;
lptr->D = scale(1.0);
lptr->biasact = scale(1.0);
lptr->patstart[TRAIN] = NULL;
lptr->currentpat[TRAIN] = NULL;
lptr->patstart[TEST] = NULL;
lptr->currentpat[TEST] = NULL;
lptr->backlayer = prevlayer;
lptr->next = NULL;
lptr->units = front;   /* connect the list of units */

/* return if this is the input layer */

if (prevlayer == NULL) return(lptr);
prevlayer->next = lptr;

/* If we are working on a deeper layer, for every node in this layer, */
/* create a linked list back to units in the previous layer. */

i = 1;
q = front;
while (q != NULL) /* do a unit */
 {    
  j = 1;            /* handle first connection */
  p = (UNIT *) prevlayer->units;
  wfront = (WTNODE *) malloc(sizeof(WTNODE));
  wttotal = wttotal + 1;
  q->wtlist = wfront;
  wprev = wfront;
  wfront->backunit = p;
  setweight(wfront,i,j,0);
  p = p->next;
  while (p != NULL) /* handle rest of connections */
   {
    j = j + 1;
    w = (WTNODE *) malloc(sizeof(WTNODE));
    wttotal = wttotal + 1;
    wprev->next = w;
    w->backunit = p;
    setweight(w,i,j,0);
    wprev = w;
    p = p->next;
   };
  j = j + 1;
  bias = (UNIT *) malloc(sizeof(UNIT));   /* create a bias unit */
  bias->oj = scale(1.0);
  bias->inuse = 1;
  bias->layernumber = nlayers;
  bias->unitnumber = 32767;           /* bias unit is unit 32767 */
  w = (WTNODE *) malloc(sizeof(WTNODE)); /* connect to end of list */
  wttotal = wttotal + 1;
  wprev->next = w;
  w->backunit = bias;
  setweight(w,n+2,i,1);
  w->next = NULL;
  q = q->next;
  i = i + 1;
 };
return(lptr);
}

#ifndef SYMMETRIC

void connect(UNIT *a,UNIT *b,WTTYPE range)
// add a connection from unit a to unit b
// connections go in increasing order

{WTNODE *wnew, *w, *wprev;
 UNIT *wunit;
 int farenough;

wnew = (WTNODE *) malloc(sizeof(WTNODE));
wttotal = wttotal + 1;
wnew->eta = dbdeta;
wnew->weight = range * (rand() & 32767) / 32768;
if ((rand() & 32767) > 16383) wnew->weight = -wnew->weight;
wnew->olddw = 0;
wnew->slope = 0;
wnew->inuse = 1;
wnew->backunit = a;
w = (WTNODE *) b->wtlist;
wprev = NULL;
wunit = (UNIT *) w->backunit;
farenough = 0;                  /* insert the weight in order */
while (w != NULL && !farenough)
 if (wunit->layernumber > a->layernumber) farenough = 1;
 else if (wunit->layernumber == a->layernumber)
  while (w != NULL && !farenough)
   {
    if (wunit->unitnumber < a->unitnumber &&
        wunit->layernumber == a->layernumber)
     {
      wprev = w;
      w = w->next;
      wunit = (UNIT *) w->backunit;
     }
    else farenough = 1;
   }
  else
   {
    wprev = w;
    w = w->next;
    wunit = (UNIT *) w->backunit;
   };
if (wprev == NULL)
 {
  wnew->next = w;
  b->wtlist = wnew;
 }
else
 {
  wnew->next = w;
  wprev->next = wnew;
 };
}
#endif

void init()
{int i,j;
 SEEDNODE *snode;

breakloop = 0;

poft = scale(0.0);     /* part of a temporal difference test */

alpha = scale(0.9);    /* gradient descent parameters */
eta = scale(0.5);
eta2 = eta;

decay = scale(0.5);    /* DBD parameters */
dbdeta = scale(0.5);
etamax = scale(30.0);
kappa = scale(0.5);
theta1 = scale(0.5);
theta2 = scale(1.0) - theta1;

mu = scale(1.75);      /* quickprop parameters */
qpdecayh = 0;
qpdecayo = 0;
qpeta = scale(0.5);
qpnoise = 0;
qpslope = '+';

classoff = 0;
classon = scale(1.0);
debugoutput = '-';
inputfile = "(none)";
bufferend = 0;
bufferptr = BUFFSIZE + 1;
ch = ' ';
copyflag = '-';
copyfilecount = 0;
deriv = 'd';
echo = '-';

extraconnect = 0;
format[0] = 0;
for(i=1;i<=MAXFORMAT-1;i++) format[i] = format[i-1] + 10;
incrementseed = '-';
informat = 'r';
initrange = scale(1.0);
kickrange = 0;
kicksize = 0;
wtinitroutine = '0';
lastadd = 0;
lastprint = 0;
lastsave = 0;
probtype = 'g';
noise = 0;
offby1 = '-';
outformat = 'r';
pagesize = 24;
pccnet = 0;
ioconnects = 0;
stmsize = 0;
recurinit = '-';
prevnpats = 0;
printrate = 10;

ringbell = '-';
saverate = MAXINT;
minimumsofar = MAXINT;
saveonminimum = '-';
seedstart = (SEEDNODE *) malloc(sizeof(SEEDNODE));
snode = (SEEDNODE *) malloc(sizeof(SEEDNODE));
seedstart->next = snode;
snode->val = 0;
srand(0);
snode->next = NULL;
#ifdef LOOKUP
{
 int i; double x,y;
 for (i=0;i<=7807;i++)
  {
   x = (double) i / 1024.0;
   y = 1.0 / (1.0 + exp(-x));
   sigmoid[i] = (WTTYPE) (y * 1024.0 + 0.5);
  };
}
#endif

for (i=TRAIN;i<=TEST;i++) for (j=TOL;j<=MAX;j++)
 {
  s[j][i].uncertain = 0;
  s[j][i].npats = 0;
  s[j][i].iterno = 0;
  s[j][i].right = 0;
  s[j][i].wrong = 0;
  s[j][i].avgerr = 0.0;
  s[j][i].pctright = 0.0;
 };
stdthresh = 0;
biasset = 0;
sysstr[0] = '\0';
summary = '+';
timestoread = 1;
toler = scale(0.1);
toloverall = 0.0;
toosmall = 0;            /* size of weights too small to use */
trigger = 0.1;
unknown = scale(0.5);
update = 'p';
up_to_date_stats = '-';
wtfilecount = 0;
wtfile = "weights";
trfiles[0] = '\0';
runningflag = '+';
wtformat = 'r';
#ifdef INTEGER
wtlimithit = 0;
#endif
wttotal = 0;
zeroderiv = 0;
maxiter = 100;

ah = 's';
ao = 's';
hbiasact = scale(1.0);
obiasact = scale(1.0);
Dh = scale(1.0);
Do = scale(1.0);

ninput = 0;
n1hidden = 0;
n2hidden = 0;
noutput = 0;

windheight = 600;
windwidth = 800;
upperleftx = 0;
upperlefty = 0;
backgroundcolor = 5;
textcolor = 2;
maxlines = 100;
currentfont = 7;

errormessages = 'b';
strcpy(tkbuf,"settext \"");
tkbufptr = 9;
}

int nonetwork()
{
if (start != NULL) return(0);
error("there is no network\n");
return(1);
}

int nopatterns()
{
if (s[TOL][TRAIN].npats != 0) return(0);
error("there are no patterns\n");
return(1);
}

void setbiasact(LAYER *layer) /* sets the bias unit ACTIVATION VALUE */
{
 UNIT *u;
 WTNODE *w;
 
 u = layer->units;
 while (u != NULL)
  {
   w = u->wtlist;
   while (w != NULL)
    {
     if (w->backunit->unitnumber == 32767)
        w->backunit->oj = layer->biasact;
     w = w->next;
    };
   u = u->next;
  };
}

void patternencode(int list)
// finds addresses of pattern values found within the net
{
 PATNODE *pn;
 PATLIST *pl;
 int i, n, finished;
 UNIT *u;
 resetpats(list);
 pl = start->currentpat[list];
 nextpat(list);
 while (pl != NULL)
  {
   pn = pl->pats;
   n = start->unitcount;
   i = 1;
   finished = 0;
   while (i <= n && !finished)
    {
     if (pn->addr != NULL && pn->addr != (WTTYPE *) -1)
      {
       u = locateunit((int) pn->layer, (int) pn->unitno);
       pn->addr = &(u->oj);
      };
     i = i + 1;
     pn++;
     if (pn->unitno == CAPXCODE || pn->unitno == CAPHCODE) finished = 1;
    };
   nextpat(list);
  };
}

void freenet(LAYER *oldstart)
{
LAYER *l, *nextl;
UNIT *u, *nextu;
WTNODE *w, *nextw;
l = oldstart;
while (l != NULL)
 {
  u = l->units;
  while (u != NULL)
   {
    w = u->wtlist;
    while (w != NULL)
     {
      nextw = w->next;
      free(w);
      w = nextw;
     };
    nextu = u->next;
    free(u);
    u = nextu;
   };
  nextl = l->next;
  free(l);
  l = nextl;
 };
}

int command(char *cs, int frominputcmd) /* process one command */
// returns QUIT for quitting the program
// returns SUCCESS for a success (default)
// returns FAILURE for a failure
{
int layerno, unitno, layer1, layer2, node1, node2, result;
int i, itemp, itemp2;
WTTYPE temp, temp2, *target;
PATLIST *pl;
REAL rtemp;
LAYER *p;
UNIT *u, *n1, *n2;
WTNODE *w;
SEEDNODE *snode, *sprev;
char *buf;

result = SUCCESS;

i = 0;
buf = &buffer[0];
while (*cs != '\0')
 {
  *buf++ = *cs++;
  i = i + 1;
 };
*buf++ = '\n';
*buf++ = '\0';
bufferend = i+2;
bufferptr = 0;
ch = ' ';


lineno = 0;
while (ch == ' ' || ch == '\n') ch = readch();
   switch (ch) {

case EOF:
popfile();
break;

case '?': parameters(stdout); break;

case '!':
i = 0;
do ch = readch(); while (ch == ' ');
if (ch == '\n')
 {
  system(sysstr);
  break;
 }
else if (ch == '"') break; /* loading string but not executing */
while ((ch != '\n' && ch != '"') && i <= 80)
 {
  sysstr[i] = ch;
  ch = readch();
  i = i + 1;
 };
if (ch == '"') sysstr[i-1] = '\0';
else
 {
  sysstr[i] = '\0';
  system(sysstr);
 };
bufferptr = bufferptr - 1;
break;

case '*': break;  /* * on a line is a comment */

case 'a':
do ch = readch(); while (ch == ' ');
if (ch == '0' || ch == '.')
 {
  bufferptr = bufferptr - 1;
  temp = rdr(GE,0.0,'a');
  if (!readerror) alpha = temp;
  goto enda;
 }
else bufferptr = bufferptr - 1;
if (nonetwork()) goto enda;
while (ch != '\n' && ch != '*')
 {
  ch = readch();
  if (ch == 'a')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == 'l' || ch == 's')
       {
        p = start;
        while (p != NULL) {p->activation = ch; p = p->next;};
        ah = ch;
        ao = ch;
       }
    else if (ch == 'h')
     {
      do ch = readch(); while (ch == ' ');
      if (ch == 'l' || ch == 's')
         {
          p = start;
          while (p != last) {p->activation = ch; p = p->next;};
          ah = ch;
         }
      else {texterror(); return(FAILURE);};
     }
    else if (ch == 'o')
     {
      do ch = readch(); while (ch == ' ');
      if (ch == 'l' || ch == 's')
         {
          last->activation = ch;
          ao = ch;
         }
      else {texterror(); return(FAILURE);};
     }
    else {texterror(); return(FAILURE);};
   }
  else if (ch == 'D')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == 'h')
     {
      temp = rdr(GT,0.0,'a');
      if (readerror) return(FAILURE);
      p = start;
      while (p != last) {p->D = temp; p = p->next;};
      Dh = temp;
     }
    else if (ch == 'o')
     {
      temp = rdr(GT,0.0,'a');
      if (readerror) return(FAILURE);
      last->D = temp;
      Do = temp;
     }
    else if (ch >= '0' && ch <= '9')
     {
      bufferptr = bufferptr - 1;
      temp = rdr(GT,0.0,'a');
      if (readerror) return(FAILURE);
      p = start;
      while (p != NULL) {p->D = temp; p = p->next;};
      Do = temp;
      Dh = temp;
     }
    else {texterror(); return(FAILURE);};
   }
  else if (ch == 'd')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == 'd' || ch == 'f' || ch == 'c' || ch == 'o') deriv = ch;
    else { texterror(); return(FAILURE);};
    if (deriv == 'o') deriv = 'c';
   }
  else if (ch == 'i')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == '+' || ch == '-') recurinit = ch;
    else {texterror(); return(FAILURE);};
   }
  else if (ch == 'u')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == 'j') ch = 'd';
#ifdef SYMMETRIC
    if (ch == 'C' || ch == 'c' || ch == 'p') update = ch;
#else
    if (ch == 'C' || ch == 'c' || ch == 'p' || ch == 'd' || ch == 'q')
       update = ch;
#endif
    else {texterror(); goto enda;};
   }
  else if (ch == '*' || ch == '\n' || ch == ' ');
  else {texterror(); return(FAILURE);};
 }
bufferptr = bufferptr - 1;
enda: break;

case 'c': {int initialize;
initialize = 0;
do ch = readch(); while (ch == ' ');
if (ch == '\n' || ch == '*')
 {if (nonetwork()) return(FAILURE); else goto clearnet;}
else if (ch == 'd')
 {
  char *newpathname;
  int status;
  newpathname = readstr();
  if (newpathname == NULL)
   {
    error("ran out of memory\n");
    return(FAILURE);
   };
  status = chdir(newpathname);
  if (status == 0) strcpy(path,newpathname);
  else
   {
    sprintf(outstr,"Change To: %s Failed\n",newpathname);
    error(outstr);
    return(FAILURE);
   };
  goto endc;
 }
else if (ch == 't') /* the copy to function */
 {
  copyfilename = readstr();
  if (copyfilename == NULL)
   {
    error("ran out of memory\n");
    return(FAILURE);
   };
  if (copy != NULL)
   {
    fflush(copy);
    fclose(copy);
   };
  copy = fopen(copyfilename,"w");
  if (copy == NULL)
   {
    sprintf(outstr,"cannot open: %s\n",copyfilename);
    return(FAILURE);
   };
  copyflag = '+';
  goto endc;
 }
else if (ch == 'i')
 {
  if (nonetwork()) return(FAILURE);
  initialize = 1;
  do ch = readch(); while (ch == ' ');
  if (ch == '\n' || ch == '*') goto clearnet;
  else
     {
      bufferptr = bufferptr - 1;
      temp2 = readreal(GE,scale(0.0),'c');
      if (readerror) return(FAILURE);
      initrange = temp2;
     };
 };

clearnet:
clear();
srand(seed);
if (initialize == 1)
 {
  kick((WTTYPE) 0,initrange);
  sprintf(outstr,"seed = %d, range = -%5.2f to +%5.2f\n",seed,
          unscale(initrange), unscale(initrange));
  pg(stdout,outstr);
 }
else {sprintf(outstr,"seed = %d\n",seed); pg(NULL,outstr);};
endc: break;};


#ifndef SYMMETRIC
case 'd':
while (ch != '\n' && ch != '*')
 {
  ch = readch();
  if (ch == 'd')
   {
    temp = rdr(GT,0.0,'d');
    if (!readerror) decay = temp;
   }
  else if (ch == 'e')
   {
    temp = rdr(GT,0.0,'d');
    if (!readerror && !nonetwork())
     {
      dbdeta = temp;
      initialize_etas(dbdeta);
     }
    else return(FAILURE);
   }
  else if (ch == 'k')
   {
    temp = rdr(GT,0.0,'d');
    if (!readerror) kappa = temp;
    else return(FAILURE);
   }
  else if (ch == 'm')
   {
    temp = rdr(GT,0.0,'d');
    if (!readerror) etamax = temp;
    else return(FAILURE);
   }
  else if (ch == 't')
   {
    temp = rdr(GE,0.0,'d');
    if (!readerror)
     {
      theta1 = temp;
      theta2 = scale(1.0) - theta1;
     }
    else return(FAILURE);
   }
  else if (ch == '*' || ch == '\n' || ch == ' ');
  else {texterror(); return(FAILURE);};
 }
bufferptr = bufferptr - 1;
endd:
break;
#endif

case 'e':
temp = rdr(GT,0.0,'e');
if (!readerror) eta = temp;
do ch = readch(); while (ch == ' ');
bufferptr = bufferptr - 1;
if (ch != '\n' && ch != '*')
 {
  temp = rdr(GT,0.0,'r');
  if (!readerror) eta2 = temp;
 }
else eta2 = eta;
break;

case 'f':
while (ch != '\n' && ch != '*')
 {
  ch = readch();
  if (ch == 'B')
   {
    itemp = 0;
    do ch = readch(); while (ch == ' ');
    while (ch >= '1' && ch <= '9')
     {
      bufferptr = bufferptr - 1;
      itemp2 = readint(format[itemp],MAXINT,'f');
      if (readerror) return(FAILURE);
      itemp = itemp + 1;
      if (itemp < MAXFORMAT) format[itemp] = itemp2;
      else error("format too long\n");
      do ch = readch(); while (ch == ' ');
     };
    if (itemp < MAXFORMAT-1)
       for (i=itemp+1;i<=MAXFORMAT-1; i++) format[i] = format[i-1] + 10;
   }
  else if (ch == 'b')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == '+' || ch == '-') ringbell = ch;
    else { texterror(); return(FAILURE);};
   }
  else if (ch == 'c')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == '+')
     {
      if (copy != NULL) fflush(copy);
      else copy = fopen("copy","w");
      if (copy == NULL)
       {sprintf(outstr,"cannot open file:  %s\n",outstr); error(outstr); };
      copyflag = '+';
     }
    else if (ch == '-')
     {
      copyflag = '-';
      if (copy != NULL)
       {
        fflush(copy);
        fclose(copy);
       }
     }
    else {texterror(); return(FAILURE);};
   }
  else if (ch == 'd')
   {/* with fd+ I can turn on a piece of debugging code (if it exists) */
    do ch = readch(); while (ch == ' ');
    if (ch == '+' || ch == '-') debugoutput = ch;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 'E')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == 'b' || ch == 't' || ch == 'w') errormessages = ch;
    else { texterror(); return(FAILURE);}
   }
  else if (ch == 'e')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == '+' || ch == '-') echo = ch;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 'i')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == 'c' || ch == 'r') informat = ch;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 'O')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == '+' || ch == '-') offby1 = ch;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 'o')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == 'a' || ch == 'c' || ch == 'r' || ch == 'e' || ch == 's')
       outformat = ch;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 'P')
   {
    itemp = readint(0,MAXINT,'f');
    if (!readerror) pagesize = itemp;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 'p')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == 'c' || ch == 'g') probtype = ch;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 'R')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == '+' || ch == '-') runningflag = ch;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 's')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == '+' || ch == '-') summary = ch;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 'u')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == '+' || ch == '-') up_to_date_stats = ch;
    else { texterror(); return(FAILURE); }
   }
  else if (ch == 'x')
   {
    temp = rdr(GT,(REAL) SKIPCODE,'f');
    if (!readerror) unknown = temp; else return(FAILURE);
   }
  else if (ch == ' ' || ch == '*' || ch == '\n');
       else { texterror(); return(FAILURE); }
 }
bufferptr = bufferptr - 1;
endf:
break;

case 'g':
{ int itemp;
  while (ch != '\n' && ch != '*')
   {
    ch = readch();
    if (ch == 'b')
     {
      itemp = readint(1,100,'g');
      if (readerror) return(FAILURE);
      backgroundcolor = itemp;
     }
    else if (ch == 'f')
     {
      itemp = readint(0,100,'g');
      if (readerror) return(FAILURE);
      currentfont = itemp;
     }
    else if (ch == 'g')
     {
      itemp = readint(1,3200,'g');
      if (readerror) return(FAILURE);
      windwidth = itemp;
      itemp = readint(1,3200,'g');
      if (readerror) return(FAILURE);
      windheight = itemp;
      itemp = readint(1,3200,'g');
      if (readerror) return(FAILURE);
      /* Why 5?  Why 22? */
      upperleftx = itemp - 5;
      itemp = readint(1,3200,'g');
      if (readerror) return(FAILURE);
      upperlefty = itemp - 22;
     }
    else if (ch == 'm')
     {
      itemp = readint(45,MAXINT,'g');
      if (readerror) return(FAILURE);
      maxlines = itemp;
     }
    else if (ch == 't')
     {
      itemp = readint(0,100,'g');
      if (readerror) return(FAILURE);
      textcolor = itemp;
     }
    else if (ch == ' ' || ch == '*' || ch == '\n');
    else texterror();
   } /* end while */
endg:
bufferptr = bufferptr - 1;
};
break;

case 'l':
if (nonetwork()) break;
layerno = readint(1,nlayers,'l'); 
if (readerror) return(FAILURE);
p = start;
for (i=2;i<=layerno;i++) p = p->next;
printoutunits(0,p,-1,0.0,(char) 0,(char) 0);
break;

/* make a network */

case 'm': {LAYER *layer; int itemp, itemp2;
nlayers = 0;
wttotal = 0;
pccnet = 0;
ioconnects = 0;
if (start != NULL)
 {
  nullpatterns(TRAIN);
  nullpatterns(TEST);
  freenet(start);
 };
ch = readch();
p = NULL;
while (ch != 'x' && ch != '\n' && ch != '*')
 {
  itemp = readint(1,MAXINT,'m');
  if (readerror)
   {
    wttotal = 0;
    start = NULL;
    goto endm;
   };
  /* check for + number for a recurrent net */
  if (nlayers == 0)
   {
    do ch = readch(); while (ch == ' ');
    if (ch != '+')
     {
      bufferptr = bufferptr - 1;
      stmsize = 0;
     }
    else
     {
      itemp2 = readint(1,MAXINT,'m');
      if (readerror)
       {
        wttotal = 0;
        start = NULL;
        goto endm;
       };
      stmsize = itemp2;
      itemp = itemp + itemp2;
     };
   };
  nlayers = nlayers + 1;
  p = mklayer(p,itemp);
  if (nlayers == 1) start = p;
  ch = readch();
  while (ch == ' ') ch = readch();
  if (ch >= '0' && ch <= '9') bufferptr = bufferptr - 1;
 };
last = p;
ninput = start->unitcount;
noutput = last->unitcount;
if (nlayers == 3 || nlayers == 4) n1hidden = start->next->unitcount;
if (nlayers == 4) n2hidden = start->next->next->unitcount;
noutput = last->unitcount;

#ifndef SYMMETRIC

if (ch == 'x')
 {
  n1 = start->units;
  while (n1 != NULL)
   {
    n2 = last->units;
    while (n2 != NULL)
     {
      connect(n1,n2,(WTTYPE) 0);
      n2 = n2->next;
     };
    n1 = n1->next;
   }
  ioconnects = 1;
 };

#endif

wtsinuse = wttotal;
bufferptr = bufferptr - 1;
nullpatterns(TRAIN);
nullpatterns(TEST);
last->activation = ao;
last->D = Do;
layer = last->backlayer;
while (layer != start)
 {
  layer->D = Dh;
  layer->activation = ah;
  layer = layer->backlayer;
 };
clear();
endm: break;};

case 'o':
do ch = readch(); while (ch == ' ');
 {
  PATNODE *targetpn;
  bufferptr = bufferptr - 1; /* unget the character */
  if (nonetwork() || nopatterns()) break;
  itemp = readint(1,s[TOL][TRAIN].npats,'o');
  if (readerror) break;
  resetpats(TRAIN);
  for (i=1;i<=itemp;i++) nextpat(TRAIN);
  /* setoutputpat(); */
  u = (UNIT *) last->units;
  pl = last->currentpat[TRAIN];
  targetpn = pl->pats;
  itemp2 = 0; /* unit counter */
  i = 1;      /* format counter */
  while (u != NULL)
   {
    sprintf(outstr,"%5.2f",unscale(targetpn->val)); pg(stdout,outstr);
    targetpn++;
    itemp2 = itemp2 + 1;
    if (format[i] == itemp2)
     {
      if (outformat == 'r') pg(stdout,"\n"); else pg(stdout," ");
      if (i < MAXFORMAT - 1) i = i + 1;
     };
    u = u->next;
   }
  pg(stdout,"\n");
 };
endo:
break;

case 'p':
do ch = readch(); while (ch == ' ');
bufferptr = bufferptr - 1;
if (ch == 'a' || ch == '\n' || ch == '*')
 {int stopprinting;
  if (nonetwork() || nopatterns()) return(FAILURE);
  stopprinting = eval(TRAIN,1);
  if (!stopprinting) printstats(stdout,TRAIN,0,s);
 }
else if (ch == 't')
 {
  PATNODE *targetpn;
  bufferptr = bufferptr + 1;
  if (nonetwork() || nopatterns()) return(FAILURE);
  itemp = readint(1,s[TOL][TRAIN].npats,'p');
  if (readerror) break;
  resetpats(TRAIN);
  for (i=1;i<=itemp;i++) nextpat(TRAIN);
  /* setoutputpat(); */
  u = (UNIT *) last->units;
  pl = last->currentpat[TRAIN];
  targetpn = pl->pats;
  itemp2 = 0; /* unit counter */
  i = 1;      /* format counter */
  while (u != NULL)
   {
    sprintf(outstr,"%5.2f",targetpn->val);
    pg(stdout,outstr);
    targetpn++;
    itemp2 = itemp2 + 1;
    if (format[i] == itemp2)
     {
      if (outformat == 'r') pg(stdout,"\n"); else pg(stdout," ");
      if (i < MAXFORMAT - 1) i = i + 1;
     };
    u = u->next;
   }
  pg(stdout,"\n");
/*
  if (nonetwork() || nopatterns()) break;
  if (testpat != 0) evalone(testpat,TRAIN,1,0);
  else error("no test pattern\n");
*/
 }
else
 {
  itemp = readint(0,s[TOL][TRAIN].npats,'p');
  if (readerror) return(FAILURE);
  if (nonetwork() || nopatterns()) return(FAILURE);
  if (itemp == 0) {eval(TRAIN,0); printstats(stdout,TRAIN,0,s);}
  else evalone(itemp,TRAIN,1,0);
 };
break;

case '.':
case '0':
case '1':
case '2':
case '3':
case '4':
case '5':
case '6':
case '7':
case '8':
case '9':
case 'x':
case '-':
case '\\': {int status;
if (nonetwork()) break;
if (ch != '\\') bufferptr = bufferptr - 1;
status = loadpat();
if (status == 1) printoutunits(0,last,-1,0.0,(char) 0,(char) 0);
else return(FAILURE);
break;}

case 'q':
ch = readch();
if (ch == '\n')
 {
  if (copy != NULL) {fflush(copy); fclose(copy);};
  Tcl_Eval(interp,"exit");
 }
#ifndef SYMMETRIC
else if (ch != 'p') texterror();
else
 {
  while (ch != '\n' && ch != '*')
   {
    ch = readch();
    if (ch == 'd')
     {REAL rtemp;
      do ch = readch(); while (ch == ' ');
      if (ch == 'h')
       {
        rtemp = readreal(GE,0.0,'q');
        if (readerror) return(FAILURE);
        qpdecayh = rtemp;
       }
      else if (ch == 'o')
       {
        rtemp = readreal(GE,0.0,'q');
        if (readerror) return(FAILURE);
        qpdecayo = rtemp;
       }
      else if (ch >= '0' && ch <= '9')
       {
        bufferptr = bufferptr - 1;
        rtemp = readreal(GE,0.0,'q');
        if (readerror) return(FAILURE);
        qpdecayh = rtemp;
        qpdecayo = rtemp;
       }
     }
    else if (ch == 'e')
     {
      temp = rdr(GT,0.0,'d');
      if (!readerror && !nonetwork()) qpeta = temp;
      else return(FAILURE);
     }
    else if (ch == 'm')
     {
      temp = rdr(GT,0.0,'d');
      if (!readerror) mu = temp;
      else return(FAILURE);
     }
    else if (ch == 'n')
     {
      temp = rdr(GE,0.0,'d');
      if (!readerror) qpnoise = temp;
      else return(FAILURE);
     }
    else if (ch == 's')
     {
      do (ch = readch()); while (ch == ' ');
      if (ch == '+' || ch == '-') qpslope = ch;
      else return(FAILURE);
     }
    else if (ch == '*' || ch == '\n' || ch == ' ');
    else {texterror(); return(FAILURE);};
   }
  };
endq:
bufferptr = bufferptr - 1;
break;
#endif

case 'r': /* r for run, rw for restore weights */
do ch = readch(); while (ch == ' ');
if (ch == '\n' || ch == '*')
 {
  if (nonetwork() || nopatterns()) goto endr;
  run(maxiter,printrate);
 }
else if (ch == 'w')
 {
  do ch = readch(); while (ch == ' ');
  bufferptr = bufferptr - 1;
  if (ch == '*' || ch == '\n') /* nothing */ ;
  else
   {
    wtfile = readstr();
    strcpy(wtfilename,wtfile);
    if (saveonminimum == '+')
     {int i, ok, temp;
      i = 0;
      while (wtfile[i] != '\0') i = i + 1;
      while (wtfile[i] != '.') i = i - 1;
      wtfile[i] = '\0';
      ok = sscanf(&wtfilename[i+1],"%d",&temp);
      if (ok == 1) wtfilecount = temp;
      else
       {
        wtfilecount = 0;
        error("weight file number error; number starts at 0\n");
       };
     };
   };
  if (nonetwork()) break; else restoreweights();
 }
else if (ch == 't')
 {
  do ch = readch(); while (ch == ' ');
  if (ch == '{') {/* nothing */}
  else
   {
    bufferptr = bufferptr - 1;
    trainfile = readstr();
    itemp = MAXINT;
    if (!pushfile(trainfile,1)) return(FAILURE);
    strcpy(trfiles,"rt ");
    strcat(trfiles,trainfile);
    strcat(trfiles,"\n");
   };
  nullpatterns(TRAIN);
  readingpattern = 1;
  itemp2 = readpats(TRAIN,'r');
  readingpattern = 0;
  s[TOL][TRAIN].npats = itemp2;
  s[MAX][TRAIN].npats = itemp2;
  sprintf(outstr,"%d training patterns read\n\n",itemp2); pg(stdout,outstr);
  goto endr;
 }
else if (ch == 'x')
 {
  if (nonetwork()) break;
  trainfile = readstr();
  if (!pushfile(trainfile,1)) return(FAILURE);
  strcat(trfiles,"rx ");
  strcat(trfiles,trainfile);
  strcat(trfiles,"\n");
  prevnpats = s[TOL][TRAIN].npats;
  findendofpats(last);
  findendofpats(start);
  readingpattern = 1;
  itemp2 = readpats(TRAIN,'r');
  sprintf(outstr,"%d patterns added\n\n",itemp2); pg(stdout,outstr);
  readingpattern = 0;
  itemp = prevnpats + itemp2;
  s[TOL][TRAIN].npats = itemp;
  s[MAX][TRAIN].npats = itemp;
  goto endr;
 }
else if (ch >= '1' && ch <= '9')
 {
  bufferptr = bufferptr - 1;
  itemp = readint(1,MAXINT,'r'); 
  if (!readerror) maxiter = itemp; else goto endr;
  itemp = readint(1,MAXINT,'r');
  if (!readerror) printrate = itemp; else goto endr;
  do ch = readch(); while (ch == ' ');
  if (ch == '"') goto endr; else bufferptr = bufferptr - 1;
  if (!nonetwork() && !nopatterns()) run(maxiter,printrate);
 }
else texterror();
endr:
break;

case 's':  /* s <int> for seed, sw <filename> for save weights */
do ch = readch(); while (ch == ' ');
if (ch == '\n' || ch == '*')
 {
  seed = clock() % 32768;
  srand(seed);
  return(SUCCESS);
 }
else if (ch == 'w')
 {
  do ch = readch(); while (ch == ' ');
  bufferptr = bufferptr - 1;
  if (ch == '*' || ch == '\n') /* nothing */ ; else wtfile = readstr();
  if (nonetwork()) break; else saveweights();
 }
else if (ch == 'e' || ch == 'a')
 {char *savefile;
  FILE *sf;
  saveweights();
  do ch = readch(); while (ch == ' ' || ch == '\n');
  bufferptr = bufferptr - 1;
  savefile = readstr();
  sf = fopen(savefile,"w");
  if (sf == NULL)
   {
    sprintf(outstr,"cannot open the file: %s\n",savefile);
    goto ends;
   };
  parameters(sf);
  fflush(sf);
  fclose(sf);
 }
else if (ch == 'i' || (ch >= '0' && ch <= '9'))
 {
  bufferptr = bufferptr - 1;
  sprev = seedstart;
  while (ch != '\n' && ch != '*')
   {
    do ch = readch(); while (ch == ' ');
    if (ch == 'i')
     {
      do ch = readch(); while (ch == ' ');
      if (ch == '+' || ch == '-') incrementseed = ch;
      else {texterror(); goto ends;};
     }
    else if (ch >= '0' && ch <= '9')
     {
      bufferptr = bufferptr - 1;
      seed = readint(0,MAXINT,'s');
      if (readerror) goto ends;
      snode = (SEEDNODE *) malloc(sizeof(SEEDNODE));
      snode->val = seed;
      snode->next = NULL;
      sprev->next = snode;
      sprev = snode;
     }
    else if (ch == '\n' || ch == '*');
    else {texterror(); goto ends;};
   };
  snode = seedstart->next;
  seed = snode->val;
 }
else if (ch == 'b')
 {WTTYPE temp; LAYER *layer; UNIT *u;
 if (nonetwork()) break;
 temp = rdr(GT,(REAL) -unscale(32767),'s');
 if (readerror) break;
 stdthresh = temp;
 layer = start->next;
 while (layer != NULL)
  {
   u = (UNIT *) layer->units;
   while (u != NULL)
    {
     w = (WTNODE *) u->wtlist;
     while (w->next != NULL) w = w->next;
#ifdef SYMMETRIC
     *(w->weight) = stdthresh;
#else
     w->weight = stdthresh;
#endif
     u = u->next;
    };
   layer = layer->next;
  };
  biasset = 1;
 }
else texterror();
ends:
endsw:
break;

case 't':
do ch = readch(); while (ch == ' ');
if (ch == 'a' || ch == '\n' || ch == '*')
 {
  if (nonetwork()) return(FAILURE);
  if (s[TOL][TEST].npats == 0) error("no test patterns\n");
  else
   {
    eval(TEST,1);
    printstats(stdout,TEST,0,s);
    bufferptr = bufferptr - 1;
   };
 }
else if (ch == 'o')
 {
  rtemp = readreal(GE,0.0,'t');
  if (readerror) break;
  toloverall = rtemp;
 }
else if (ch == 'f')
 {
  do ch = readch(); while (ch == ' ');
  bufferptr = bufferptr - 1;
  testfile = readstr();
  nullpatterns(TEST);
  if (!pushfile(testfile,1)) goto endt;
  readingpattern = 1;
  itemp = readpats(TEST,'t');
  readingpattern = 0;
  sprintf(outstr,"%d test patterns read\n",itemp); pg(stdout,outstr);
  s[TOL][TEST].npats = itemp;
  s[MAX][TEST].npats = itemp;
 }
else if (ch == 'r')
 {
  do ch = readch(); while (ch == ' ');
  if (ch == 'p') itemp2 = 1;
  else {itemp2 = 0; bufferptr = bufferptr - 1;};
  itemp = readint(1,s[TOL][TEST].npats,'t');
  if (!readerror) evalr(TEST,itemp2,itemp);
 }
else if (ch == 't')
 {
  PATNODE *targetpn;
  if (nonetwork() || nopatterns()) return(FAILURE);
  itemp = readint(1,s[TOL][TEST].npats,'t');
  if (readerror) return(FAILURE);
  resetpats(TEST);
  for (i=1;i<=itemp;i++) nextpat(TEST);
  /* setoutputpat(); */
  u = (UNIT *) last->units;
  pl = last->currentpat[TEST];
  targetpn = pl->pats;
  itemp2 = 0; /* unit counter */
  i = 1;      /* format counter */
  while (u != NULL)
   {
    sprintf(outstr,"%5.2f",targetpn->val);
    pg(stdout,outstr);
    targetpn++;
    itemp2 = itemp2 + 1;
    if (format[i] == itemp2)
     {
      if (outformat == 'r') pg(stdout,"\n"); else pg(stdout," ");
      if (i < MAXFORMAT - 1) i = i + 1;
     };
    u = u->next;
   }
  pg(stdout,"\n");
 }
else
 {
  bufferptr = bufferptr - 1;
  rtemp = readreal(GE,0.0,'t');
  if (readerror) break;
  else if (rtemp > 0 && rtemp < 1.0) toler = scale(rtemp);
  else
   {
    bufferptr = bufferptr - 1;
    if (s[TOL][TEST].npats == 0)
     {
      error("there is no test set\n");
      return(FAILURE);
     };
    if (rtemp == 0.0) {eval(TEST,0); printstats(stdout,TEST,0,s);}
    else
     {
      itemp = (int) rtemp;
      if (itemp > s[TOL][TEST].npats)
       error("not that many patterns in the test set\n");
      else evalone(itemp,TEST,1,0);
     };
   };
 };
endt: break;

case 'w':
{int i;
 if (nonetwork()) break;
 layerno = readint(2,nlayers,'w');
 if (readerror) return(FAILURE);
 unitno = readint(1,MAXINT,'w');
 if (readerror) return(FAILURE);
 u = locateunit(layerno,unitno);
 if (u != NULL) printweights(u,layerno);
 break;
};

#ifdef OLDSTUFF
case 'g': /* a maze (game) problem for temporal difference */
{int ngames, nlevels, ndoors;
 ngames = readint(1,10,'x');
 nlevels = readint(1,10,'x');
 ndoors = readint(1,10,'x');
 maze(ngames,nlevels,ndoors);
 ch = ' ';
 break;
};
#endif

case 'z': /* reads in non-user set (hidden) parameters */
{REAL rtemp; int itemp;
 rtemp = readreal(GE,0.0,'z');
 if (readerror) return(FAILURE);
 minimumsofar = rtemp;  /* minimum test set error */
};
break;

default: texterror(); break;
      }; // end switch
if (frominputcmd == 0)
 {
  bufferptr = -1;
  lineno = 0;
 };
return(result);
}

int inputcmd()
{
 int i, status, ch, nread;
 char str[100];

bufferptr = 0; bufferend = 1 + strlen(buffer);
ch = readch();  /* reads the i in the FIRST position */
if (ch != 'i')
 {
  error("program landed in inputcmd without an i command\n");
  return(FAILURE);
 };
ch = readch();
if (ch != '\n')
 {
  inputfile = readstr();
//  ch = readch();
//  if (ch == '"') return(SUCCESS); else bufferptr = bufferptr - 1;
 };
status = pushfile(inputfile,1);
if (status == 0)
 {
  printf(str,"Cannot open the file:  %s",inputfile);
  error(str);
  return(FAILURE);
 };

/* now reading from a new file */
 ch = fgetc(dataf);
 while (ch != EOF && filestackptr > 0)
  {
   i = 0;
   while (ch != '\n' && i < BUFFSIZE-2)
    {
     buffer[i] = ch;
     i = i + 1;
     ch = fgetc(dataf);
    };
   buffer[i] = '\n';
   i = i + 1;
   buffer[i] = '\0';
   bufferptr = 0;
   bufferend = i;
   if (echo == '+') pg(NULL,buffer);
   if (i > 1) /* if its more than a carriage return */
    {
//printf("inputcmd line: %s\n",buffer);
     status = command(buffer,1);
     if (status == QUIT || status == FAILURE)
      {
       while (filestackptr > 0) popfile();
//       if (status == FAILURE) pg(NULL,">");
       return(status);
      };
    };
   if (filestackptr > 0) ch = fgetc(dataf); else ch = EOF;
   while (ch == EOF && filestackptr > 0)
    {
     popfile();
     if (filestackptr > 0) ch = fgetc(dataf); else ch = EOF;
    };
  };
 return(1);
}

int sendpmenu(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set informat %c",informat);
 Tcl_Eval(interp,str);
 sprintf(str,"set probtype %c",probtype);
 Tcl_Eval(interp,str);
}

int sendimenu(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set wtfilename %s",wtfilename);
 Tcl_Eval(interp,str);
 sprintf(str,"set echo %c",echo);
 Tcl_Eval(interp,str);
 sprintf(str,"set copyflag %c",copyflag);
 Tcl_Eval(interp,str);
 sprintf(str,"set informat %c",informat);
 Tcl_Eval(interp,str);
 sprintf(str,"set probtype %c",probtype);
 Tcl_Eval(interp,str);
 sprintf(str,"set unknown %f",unknown);
 Tcl_Eval(interp,str);
}

int sendfmenu(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 int i;
 char buf[10];
 char str[BUFFSIZE];
 sprintf(str,"set copyfilename %s",copyfilename);
 Tcl_Eval(interp,str);
 sprintf(str,"set echo %c",echo);
 Tcl_Eval(interp,str);
 sprintf(str,"set copyflag %c",copyflag);
 Tcl_Eval(interp,str);
 sprintf(str,"set informat %c",informat);
 Tcl_Eval(interp,str);
 sprintf(str,"set errormessages %c",errormessages);
 Tcl_Eval(interp,str);
 sprintf(str,"set outformat %c",outformat);
 Tcl_Eval(interp,str);
/* offby1 is fairly useless, discontinue
 sprintf(str,"set offby1 %c",offby1);
 Tcl_Eval(interp,str);
*/
 sprintf(str,"set probtype %c",probtype);
 Tcl_Eval(interp,str);
 sprintf(str,"set runningflag %c",runningflag);
 Tcl_Eval(interp,str);
 sprintf(str,"set summary %c",summary);
 Tcl_Eval(interp,str);
 sprintf(str,"set up_to_date_stats %c",up_to_date_stats);
 Tcl_Eval(interp,str);
 sprintf(str,"set pagesize %d",pagesize);
 Tcl_Eval(interp,str);
 sprintf(str,"set unknown %f",unknown);
 Tcl_Eval(interp,str);
 strcpy(str,"set format \"");
 for (i=1;i<MAXFORMAT;i++)
    {
     sprintf(buf," %1d",format[i]);
     strcat(str,buf);
    };
 strcat(str,"\"");
 Tcl_Eval(interp,str);
}

int sendomenu(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set wtfilename %s",wtfilename);
 Tcl_Eval(interp,str);
 sprintf(str,"set copyfilename %s",copyfilename);
 Tcl_Eval(interp,str);
 sprintf(str,"set echo %c",echo);
 Tcl_Eval(interp,str);
 sprintf(str,"set copyflag %c",copyflag);
 Tcl_Eval(interp,str);
}

int sendwts(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set wtfilename %s",wtfilename);
 Tcl_Eval(interp,str);
}

int sendtrain(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set seed %d",seed);
 Tcl_Eval(interp,str);
 sprintf(str,"set maxiter %d",maxiter);
 Tcl_Eval(interp,str);
 sprintf(str,"set printrate %d",printrate);
 Tcl_Eval(interp,str);
 sprintf(str,"set pagesize %d",pagesize);
 Tcl_Eval(interp,str);
 sprintf(str,"set initrange %6.3f",initrange);
 Tcl_Eval(interp,str);
 sprintf(str,"set toler %6.3f",toler);
 Tcl_Eval(interp,str);
}

int sendalgor(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set ah %c",ah);
 Tcl_Eval(interp,str);
 sprintf(str,"set ao %c",ao);
 Tcl_Eval(interp,str);
 sprintf(str,"set deriv %c",deriv);
 Tcl_Eval(interp,str);
 sprintf(str,"set update %c",update);
 Tcl_Eval(interp,str);
 sprintf(str,"set toler %5.2f",toler);
 Tcl_Eval(interp,str);
}

int senddbd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set decay %8.6f",decay);
 Tcl_Eval(interp,str);
 sprintf(str,"set dbdeta %8.6f",dbdeta);
 Tcl_Eval(interp,str);
 sprintf(str,"set kappa %7.4f",kappa);
 Tcl_Eval(interp,str);
 sprintf(str,"set etamax %7.4f",etamax);
 Tcl_Eval(interp,str);
 sprintf(str,"set theta1 %7.4f",theta1);
 Tcl_Eval(interp,str);
 sprintf(str,"set update %c",update);
 Tcl_Eval(interp,str);
}

int sendqp(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set qpdecayh %8.6f",qpdecayh);
 Tcl_Eval(interp,str);
 sprintf(str,"set qpdecayo %8.6f",qpdecayo);
 Tcl_Eval(interp,str);
 sprintf(str,"set qpeta %7.4f",qpeta);
 Tcl_Eval(interp,str);
 sprintf(str,"set mu %7.4f",mu);
 Tcl_Eval(interp,str);
 sprintf(str,"set qpslope %c",qpslope);
 Tcl_Eval(interp,str);
 sprintf(str,"set update %c",update);
 Tcl_Eval(interp,str);
}

int sendgrad(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set eta %7.4f",eta);
 Tcl_Eval(interp,str);
 sprintf(str,"set eta2 %7.4f",eta2);
 Tcl_Eval(interp,str);
 sprintf(str,"set alpha %7.4f",alpha);
 Tcl_Eval(interp,str);
 sprintf(str,"set update %c",update);
 Tcl_Eval(interp,str);
}

int sendgui(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set windwidth %d",windwidth);
 Tcl_Eval(interp,str);
 sprintf(str,"set windheight %d",windheight);
 Tcl_Eval(interp,str);
 sprintf(str,"set upperleftx %d",upperleftx);
 Tcl_Eval(interp,str);
 sprintf(str,"set upperlefty %d",upperlefty);
 Tcl_Eval(interp,str);
 sprintf(str,"set backgroundcolor %d",backgroundcolor);
 Tcl_Eval(interp,str);
 sprintf(str,"set textcolor %d",textcolor);
 Tcl_Eval(interp,str);
 sprintf(str,"set currentfont %d",currentfont);
 Tcl_Eval(interp,str);
 sprintf(str,"set maxlines %d",maxlines);
 Tcl_Eval(interp,str);
 sprintf(str,"set tkmaxlines %d",maxlines);
 Tcl_Eval(interp,str);
 sprintf(str,"set pagesize %d",pagesize);
 Tcl_Eval(interp,str);
 sprintf(str,"set cmdfilename %s",cmdfilename);
 Tcl_Eval(interp,str);
}

int sendnetsize(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 char str[BUFFSIZE];
 sprintf(str,"set nlayers %d",nlayers);
 Tcl_Eval(interp,str);
 sprintf(str,"set outformat %c",outformat);
 Tcl_Eval(interp,str);
 sprintf(str,"set wttotal %d",wttotal);
 Tcl_Eval(interp,str);
 sprintf(str,"set ninput %d",ninput);
 Tcl_Eval(interp,str);
 sprintf(str,"set n1hidden %d",n1hidden);
 Tcl_Eval(interp,str);
 sprintf(str,"set n2hidden %d",n2hidden);
 Tcl_Eval(interp,str);
 sprintf(str,"set noutput %d",noutput);
 Tcl_Eval(interp,str);
 sprintf(str,"set stmsize %d",stmsize);
 Tcl_Eval(interp,str);
 if (ioconnects == 1) Tcl_Eval(interp,"set ioconnects Yes");
 else Tcl_Eval(interp,"set ioconnects No");
}

int stopcmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
breakloop = 1;
return(TCL_OK);
}

// for processing all the normal commands (not more dialog box results)
int ccmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
 int i;
 char cmd[BUFFSIZE], str[BUFFSIZE];
 strcpy(cmd,argv[1]);
 sprintf(str,"settext \"%s\"",cmd);
 Tcl_Eval(interp,str);
 if (copy != NULL) fprintf(copy,"%s\n",cmd);
 sigint = 0;
 setjmp(endofccmd);
 if (sigint == 1) return(TCL_OK);
 i = 0;
 while (cmd[i] == ' ' && i < (BUFFSIZE - 1)) i = i + 1;
 if (cmd[i] == 'i')
  {
   strcpy(buffer,&cmd[i]);
   inputcmd();
  }
 else command(cmd,0);
 return(TCL_OK);
}

void main(int argc, char *argv[])
{
int code; // return value from Tcl/Tk
setbuf(stdout,NULL);  /* set unbuffered output */
filestackptr = 0;
filestack[0] = stdin;
dataf = stdin;
datafile = NULL;
emptystring = '\0';
trainfile = &emptystring;
testfile = &emptystring;
init();
lineno = 0;

 interp = Tcl_CreateInterp();

 if (Tcl_Init(interp) == TCL_ERROR)
    {
     printf("Tcl_Init failed: %s\n", interp->result);
     exit(1);
    };

 if (Tk_Init(interp) == TCL_ERROR)
    {
     printf("Tk_Init failed: %s\n", interp->result);
     exit(1);
    };

 Tcl_CreateCommand(interp, "stopcmd",stopcmd,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "ccmd",ccmd,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendnetsize",sendnetsize,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendgui",sendgui,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendgrad",sendgrad,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendqp",sendqp,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "senddbd",senddbd,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendalgor",sendalgor,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendtrain",sendtrain,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendwts",sendwts,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendomenu",sendomenu,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendfmenu",sendfmenu,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendimenu",sendimenu,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);
 Tcl_CreateCommand(interp, "sendpmenu",sendimenu,(ClientData) NULL,(Tcl_CmdDeleteProc *)NULL);

if (argc > 1)
 {int n;
  char buf[25];
  strcpy(buf,argv[1]);
  if (buf[0] == '-' && buf[1] == 'g')
   {
    strcpy(buf,argv[2]);
    if (buf[0] >= '1' && buf[0] <= '9')
     {
      n = sscanf(buf,"%dx%d%d%d",&windwidth,&windheight,&upperleftx,&upperlefty);
      if (n != 4) {printf("bad geometry\n"); exit(1);};
     }
    else if (buf[0] == '+')
     {
      n = sscanf(buf,"%d%d",&upperleftx,&upperlefty);
      if (n != 2) {printf ("bad geometry\n"); exit(1);};
     }
    else
     {
      printf ("missing geometry\n");
      exit;
     };
    if (argc == 4)
     {
      datafile = argv[3];
      strcpy(cmdfilename,argv[3]);
     };
   }
  else
   {
    datafile = argv[1];
    strcpy(cmdfilename,argv[1]);
   };
 };
if (datafile == NULL) strcpy(cmdfilename,"//");
else
 {FILE *g;
  int n, tw, th, tx, ty, itemp;
  g = fopen(datafile,"r");
  if (g == NULL)
   {
    printf("cannot open: %s\n",datafile);
    exit(1);
   };
  ch = fgetc(g);
  while (ch != 'g' && ch != '\n') ch = fgetc(g);
  if (ch == 'g')
   {
    while (ch != '\n' && ch != '*')
     {
      ch = fgetc(g);
      if (ch == 'b')
       {
        n = fscanf(g,"%d",&itemp);
        if (n != 1 || itemp < 1 || itemp > 100) goto endg;
        backgroundcolor = itemp;
       }
      else if (ch == 'f')
       {
        n = fscanf(g,"%d",&itemp);
        if (n != 1 || itemp < 1 || itemp > 100) goto endg;
        currentfont = itemp;
       }
      else if (ch == 'g')
       {
        n = fscanf(g,"%d %d %d %d",&tw,&th,&tx,&ty);
        if (n == 4)
         {
          if (tw > 0) windwidth = tw;
          else printf("bad window width\n");
          if (th > 0) windheight = th;
          else printf("bad window height\n");
          /* why is it off by 5 and 22? */
          if (tx > 0) upperleftx = tx - 5;
          else printf("bad window x position\n");
          if (ty > 0) upperlefty = ty - 22;
          else printf("bad window y position\n");
         }; 
       }
      else if (ch == 'm')
       {
        n = fscanf(g,"%d",&itemp);
        if (n != 1 || itemp < 45) goto endg;
        maxlines = itemp;
       }
      else if (ch == 't')
       {
        n = fscanf(g,"%d",&itemp);
        if (n != 1 || itemp < 1 || itemp > 100) goto endg;
        textcolor = itemp;
       }
      else if (ch == ' ' || ch == '*' || ch == '\n');
      else texterror();
     } /* end while */
   }
  endg: fclose(g);
 };

code = Tcl_EvalFile(interp, "bp.tcl"); 
if (code != TCL_OK)
 {
  printf("error number from Tcl_EvalFile: %d\n",code);
  printf("%s\n",interp->result);
  exit(1);
 };
Tk_MainLoop();
}
